@@ -1,5 +1,5 @@
 /* This file is part of Mailfromd.
-   Copyright (C) 2005-2020 Sergey Poznyakoff
+   Copyright (C) 2005-2024 Sergey Poznyakoff
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -36,6 +36,7 @@
 #include <mailutils/daemon.h>
 
 #include "mailfromd.h"
+#include "prog.h"
 #include "srvman.h"
 #include "srvcfg.h"
 
@@ -47,7 +48,7 @@ static int ctx_setreply(void *data, char *code, char *xcode, char *message);
 
 struct message_data {
 	eval_environ_t env;        /* Evaluation environment */
-	char *helostr;             /* Domain name obtained in HELO phase */   
+	char *helostr;             /* Domain name obtained in HELO phase */
 	char msgid[64];            /* Message ID */
 };
 static struct message_data *test_message_data;
@@ -70,15 +71,15 @@ priv_get(SMFICTX *ctx)
 		md = test_message_data;
 	else
 		md = (struct message_data*) gacopyz_getpriv(ctx);
-		
+
 	if (!md) {
 		md = malloc(sizeof(*md));
-		if (!md) 
+		if (!md)
 			mu_error(_("not enough memory"));
 		else {
 			milter_sockaddr_t addr;
 			socklen_t len = sizeof(addr);
-		
+
 			md->env = create_environment(ctx,
 						     ctx_getsym,
 						     ctx_setreply,
@@ -94,8 +95,8 @@ priv_get(SMFICTX *ctx)
 			set_milter_server_id(md->env, gacopyz_getclosure(ctx));
 			if (gacopyz_client_sockname(ctx, &addr, &len) == 0)
 				set_milter_client_address(md->env, &addr, len);
-			
-			xeval(md->env, smtp_state_begin);
+
+			xeval_begin(md->env);
 		}
 	}
 	if (!md->msgid[0]) {
@@ -140,7 +141,7 @@ ctx_getsym(void *data, const char *str)
 static int
 ctx_setreply(void *data, char *code, char *xcode, char *message)
 {
-	if (code) 
+	if (code)
 		return gacopyz_setreply(data, code, xcode, message);
 	return 0;
 }
@@ -170,7 +171,7 @@ capture_from(eval_environ_t env, const char *str)
 		char datebuf[26];
 
 		env_capture_start(env);
-		
+
 		t = time(NULL);
 		tm = localtime(&t);
 		mu_strftime(datebuf, sizeof datebuf,
@@ -203,7 +204,7 @@ static void
 capture_eom(eval_environ_t env)
 {
 }
-	
+
 
 /* Cleanup functions */
 void
@@ -228,13 +229,20 @@ int
 xeval(eval_environ_t env, enum smtp_state tag)
 {
 	int rc;
-	
-	rc = eval_environment(env, entry_point[tag]);
-	if (rc) 
-		mu_error(_("execution of the filter program was not finished"));
-	if (tag == smtp_state_begin)
-		env_save_catches(env);
 
+	env_make_frame(env);
+	set_milter_state(env, tag);
+	rc = eval_environment(env, entry_point[tag]);
+	if (rc)
+		mu_error(_("execution of the filter program was not finished"));
+	return rc;
+}
+
+int
+xeval_begin(eval_environ_t env)
+{
+	int rc = xeval(env, smtp_state_begin);
+	env_save_catches(env);
 	return rc;
 }
 
@@ -265,62 +273,64 @@ mlfi_connect(SMFICTX *ctx, char *hostname, _SOCK_ADDR *hostaddr)
 	int port;
 	char *addrstr;
 	int family;
-	
+
 	if (!hostaddr) {
 		family = MFAM_STDIO;
 		port = 0;
 		addrstr = "";
-	} else switch (hostaddr->sa.sa_family) {
-	case PF_INET:
-		family = MFAM_INET;
-		port = ntohs(hostaddr->sin.sin_port);
-		addrstr = inet_ntoa(hostaddr->sin.sin_addr);
-		break;
+		env_save_socket_family(md->env, AF_INET);
+	} else {
+		switch (hostaddr->sa.sa_family) {
+		case PF_INET:
+			family = MFAM_INET;
+			port = ntohs(hostaddr->sin.sin_port);
+			addrstr = inet_ntoa(hostaddr->sin.sin_addr);
+			break;
 
-#ifdef GACOPYZ_IPV6
-	case PF_INET6: {
-		char host[NI_MAXHOST];
+		case PF_INET6: {
+			char host[NI_MAXHOST];
 
-		family = MFAM_INET6;
-		port = ntohs(hostaddr->sin6.sin6_port);
-		if (getnameinfo(&hostaddr->sa, sizeof(hostaddr->sin6),
-				host, sizeof host,
-				NULL, 0,
-				NI_NUMERICHOST) == 0)
-			addrstr = host;
-		else
-			addrstr = "[unresolved]";
-		break;
+			family = MFAM_INET6;
+			port = ntohs(hostaddr->sin6.sin6_port);
+			if (getnameinfo(&hostaddr->sa, sizeof(hostaddr->sin6),
+					host, sizeof host,
+					NULL, 0,
+					NI_NUMERICHOST) == 0)
+				addrstr = host;
+			else
+				addrstr = "[unresolved]";
+			break;
+		}
+
+		case PF_UNIX:
+			family = MFAM_UNIX;
+			port = 0;
+			addrstr = hostaddr->sunix.sun_path;
+			break;
+
+		default:
+			mu_error(_("mlfi_connect: unsupported address family: %d"),
+				 hostaddr->sa.sa_family);
+			gacopyz_setreply(ctx, "410", NULL,
+					 "Local configuration error; please try again later");
+			return SMFIS_TEMPFAIL;
+		}
+		env_save_socket_family(md->env, hostaddr->sa.sa_family);
 	}
-#endif
-		
-	case PF_UNIX:
-		family = MFAM_UNIX;
-		port = 0;
-		addrstr = hostaddr->sunix.sun_path;
-		break;
 
-	default:
-		mu_error(_("mlfi_connect: unsupported address family: %d"),
-			 hostaddr->sa.sa_family);
-		gacopyz_setreply(ctx, "410", NULL,
-				 "Local configuration error; please try again later");
-		return SMFIS_TEMPFAIL;
-	}
-	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, 
-              ("Processing xxfi_connect: %s, %d, %s, %u",
-	        hostname, family, addrstr, port));
+	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7,
+		 ("Processing xxfi_connect: %s, %d, %s, %u",
+		  hostname, family, addrstr, port));
 	mf_proctitle_format("Connect: %s, %d, %s, %u",
 			    hostname, family, addrstr, port);
-	
+
 	env_init(md->env);
 	env_push_string(md->env, addrstr);
 	env_push_number(md->env, port);
 	env_push_number(md->env, family);
 	env_push_string(md->env, hostname);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_connect);
-	env_leave_frame(md->env, 4);
+	adjust_stack(md->env, 4);
 	return status;
 }
 
@@ -335,13 +345,12 @@ mlfi_helo(SMFICTX *ctx, char *helohost)
 	mf_proctitle_format("%sHELO %s", md->msgid, helohost);
 
 	if (md->helostr)
-	  free (md->helostr);
+	  free(md->helostr);
 	md->helostr = mu_strdup(helohost);
 	env_init(md->env);
 	env_push_string(md->env, md->helostr);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_helo);
-	env_leave_frame(md->env, 1);
+	adjust_stack(md->env, 1);
 	return status;
 }
 
@@ -364,18 +373,17 @@ mlfi_envfrom(SMFICTX *ctx, char **argv)
 	struct message_data *md = priv_get(ctx);
 	char *p = concat_args(argv + 1);
 
-	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, 
-              ("Processing xxfi_envfrom: %s %s", argv[0], p));
+	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7,
+	      ("Processing xxfi_envfrom: %s %s", argv[0], p));
 	mf_proctitle_format("%sMAIL FROM %s %s", md->msgid, argv[0], p);
-	
+
 	env_init(md->env);
 	capture_from(md->env, argv[0]);
 	env_push_string(md->env, p);
 	free(p);
 	env_push_string(md->env, argv[0]);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_envfrom);
-	env_leave_frame(md->env, 2);
+	adjust_stack(md->env, 2);
 	return status;
 }
 
@@ -383,20 +391,19 @@ sfsistat
 mlfi_envrcpt(SMFICTX *ctx, char ** argv)
 {
 	sfsistat status;
-        struct message_data *md = priv_get(ctx);
+	struct message_data *md = priv_get(ctx);
 	char *p = concat_args(argv + 1);
 
-	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, 
+	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7,
 		 ("Processing xxfi_envrcpt: %s %s", argv[0], p));
 	mf_proctitle_format("%sRCPT TO %s %s", md->msgid, argv[0], p);
 	env_init(md->env);
 	env_push_string(md->env, p);
 	free(p);
 	env_push_string(md->env, argv[0]);
-	env_make_frame(md->env);
 	incr_rcpt_count(md->env);
 	status = mlfi_eval(ctx, smtp_state_envrcpt);
-	env_leave_frame(md->env, 2);
+	adjust_stack(md->env, 2);
 	return status;
 }
 
@@ -409,9 +416,8 @@ mlfi_data(SMFICTX *ctx)
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, ("Processing xxfi_data:"));
 	mf_proctitle_format("%sDATA", md->msgid);
 	env_init(md->env);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_data);
-	env_leave_frame(md->env, 1);
+	adjust_stack(md->env, 1);
 	return status;
 }
 
@@ -428,9 +434,8 @@ mlfi_header(SMFICTX *ctx, char *headerf, char *headerv)
 	capture_header(md->env, headerf, headerv);
 	env_push_string(md->env, headerv);
 	env_push_string(md->env, headerf);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_header);
-	env_leave_frame(md->env, 2);
+	adjust_stack(md->env, 2);
 	return status;
 }
 
@@ -438,14 +443,13 @@ sfsistat
 mlfi_eoh(SMFICTX *ctx)
 {
 	sfsistat status;
-        struct message_data *md = priv_get(ctx);
+	struct message_data *md = priv_get(ctx);
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, ("Processing xxfi_eoh"));
 	mf_proctitle_format("%sEOH", md->msgid);
 	env_init(md->env);
 	capture_eoh(md->env);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_eoh);
-	env_leave_frame(md->env, 0);
+	adjust_stack(md->env, 0);
 	return status;
 }
 
@@ -454,7 +458,7 @@ mlfi_body(SMFICTX *ctx, unsigned char *bodyp, size_t len)
 {
 	sfsistat status;
 	struct message_data *md = priv_get(ctx);
-	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, 
+	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7,
 		 ("Processing xxfi_body: %lu", (unsigned long) len));
 	mf_proctitle_format("%sBODY", md->msgid);
 	env_init(md->env);
@@ -464,9 +468,8 @@ mlfi_body(SMFICTX *ctx, unsigned char *bodyp, size_t len)
 	   User can then convert it to string using the body_string() call.
 	 */
 	env_push_pointer(md->env, bodyp);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_body);
-	env_leave_frame(md->env, 2);
+	adjust_stack(md->env, 2);
 	return status;
 }
 
@@ -495,9 +498,9 @@ xlate_and_replace_body(SMFICTX *ctx, const char *value, size_t size)
 	}
 
 	while (size) {
-	        size_t n = mem_search(value, '\n', size);
+		size_t n = mem_search(value, '\n', size);
 		size_t off;
-		
+
 		if (value[n] == '\n') {
 			off = n + 1;
 			if (n > 0 && value[n-1] == '\r')
@@ -511,7 +514,7 @@ xlate_and_replace_body(SMFICTX *ctx, const char *value, size_t size)
 				 mu_strerror(rc));
 			break;
 		}
-		
+
 		value += off;
 		size -= off;
 	}
@@ -554,7 +557,7 @@ run_msgmod(void *item, void *data)
 {
 	struct msgmod_closure *hdr = item;
 	SMFICTX *ctx = data;
-	
+
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE6,
 		 ("%s %s: %s %u",
 		  msgmod_opcode_str(hdr->opcode),
@@ -564,7 +567,7 @@ run_msgmod(void *item, void *data)
 	case header_add:
 		gacopyz_add_header(ctx, hdr->name, hdr->value);
 		break;
-		
+
 	case header_replace:
 		gacopyz_change_header(ctx, hdr->idx, hdr->name, hdr->value);
 		break;
@@ -576,7 +579,7 @@ run_msgmod(void *item, void *data)
 	case header_insert:
 		gacopyz_insert_header(ctx, hdr->idx, hdr->name, hdr->value);
 		break;
-		
+
 	case rcpt_add:
 		gacopyz_add_rcpt(ctx, hdr->name);
 		break;
@@ -584,7 +587,7 @@ run_msgmod(void *item, void *data)
 	case rcpt_delete:
 		gacopyz_del_rcpt(ctx, hdr->name);
 		break;
-		
+
 	case quarantine:
 		gacopyz_quarantine(ctx, hdr->name);
 		break;
@@ -601,11 +604,11 @@ run_msgmod(void *item, void *data)
 			break;
 		}
 		break;
-		
+
 	case set_from:
 		gacopyz_chgfrom(ctx, hdr->name, hdr->value);
 		break;
-		
+
 	default:
 		abort();
 	}
@@ -615,7 +618,7 @@ run_msgmod(void *item, void *data)
 sfsistat
 mlfi_eom(SMFICTX *ctx)
 {
-        sfsistat status;
+	sfsistat status;
 	struct message_data *md = priv_get(ctx);
 
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, ("Processing xxfi_eom"));
@@ -623,16 +626,15 @@ mlfi_eom(SMFICTX *ctx)
 
 	env_init(md->env);
 	capture_eom(md->env);
-	env_make_frame(md->env);
 	status = mlfi_eval(ctx, smtp_state_eom);
-	env_leave_frame(md->env, 0);
+	adjust_stack(md->env, 0);
 	if ((status == SMFIS_ACCEPT || status == SMFIS_CONTINUE)
 	    && env_msgmod_count(md->env) > 0) {
 		mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE6,
 			 ("flushing message modification queue"));
 		env_msgmod_apply(md->env, run_msgmod, ctx);
 	}
-	
+
 	mf_proctitle_format("%sfinished", md->msgid);
 	clear_rcpt_count(md->env);
 	env_msgmod_clear(md->env);
@@ -647,14 +649,14 @@ mlfi_abort(SMFICTX *ctx)
 
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, ("Abort"));
 	mf_proctitle_format("%saborting", md->msgid);
-        env_msgmod_clear(md->env);
-        md->msgid[0] = 0;
+	env_msgmod_clear(md->env);
+	md->msgid[0] = 0;
 	/* Note: the value of helostr is not discarded: RFC 2822,
 	   section 4.1.1.5 states that RSET issued immediately
 	   after EHLO is effectively equivalent to a NOOP. */
 	env_init_dataseg(md->env);
 	return SMFIS_CONTINUE;
-}		
+}
 
 sfsistat
 mlfi_close(SMFICTX *ctx)
@@ -662,7 +664,7 @@ mlfi_close(SMFICTX *ctx)
 	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, ("Close"));
 	mf_proctitle_format("closing");
 	filter_cleanup(ctx);
-	return SMFIS_CONTINUE; 
+	return SMFIS_CONTINUE;
 }
 
 static int
@@ -674,8 +676,6 @@ child_start()
 	return 0;
 }
 
-enum smtp_state first_used_state = smtp_state_end;
-
 sfsistat
 mlfi_negotiate(SMFICTX *ctx,
 	       unsigned long mta_actions,
@@ -689,19 +689,20 @@ mlfi_negotiate(SMFICTX *ctx,
 {
 	enum gacopyz_stage i;
 
-	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7, 
+	mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE7,
 		 ("Negotiate: mta_actions=%#lx, mta_capa=%#lx, "
 		  "filter_actions=%#lx, filter_capa=%#lx",
 		  mta_actions, mta_capa, *filter_actions, *filter_capa));
 
-	if (first_used_state > smtp_state_first
-	    && first_used_state < smtp_state_end) {
-		mu_debug(MF_SOURCE_ENGINE, MU_DEBUG_TRACE6, 
-			 ("Registering \"i\" macro for state %s",
-			  state_to_string(first_used_state)));
-		register_macro(first_used_state, "i"); 
-	}
-	
+	/*
+	 * First entered sate is connect, because smfilter.xxfi_connect is
+	 * always initialized.  This means "begin" handler (if defined) will
+	 * be run from this state.  Traditionally the "i" macro was
+	 * available in all handlers, therefore it must be registered for
+	 * that state.
+	 */
+	register_macro(smtp_state_connect, "i");
+
 	for (i = 0; i < gacopyz_stage_max; i++) {
 		char *str = get_stage_macro_string(i);
 		gacopyz_setsymlist(ctx, i, str);
@@ -712,16 +713,16 @@ mlfi_negotiate(SMFICTX *ctx,
 
 static struct smfiDesc smfilter =
 {
-	"MailfromFilter",	
-	SMFI_VERSION,	
+	"MailfromFilter",
+	SMFI_VERSION,
 	SMFI_V2_ACTS|SMFIF_CHGFROM, /* FIXME: Add flags as needed */
-	NULL,	        /* connection info filter */
-	NULL,	        /* SMTP HELO command filter */
-	NULL,	        /* envelope sender filter */
-	NULL,	        /* envelope recipient filter */
-	NULL,	        /* header filter */
+	mlfi_connect,	/* connection info filter */
+	NULL,		/* SMTP HELO command filter */
+	NULL,		/* envelope sender filter */
+	NULL,		/* envelope recipient filter */
+	NULL,		/* header filter */
 	NULL,           /* end of header */
-	NULL,	        /* body block filter */
+	NULL,		/* body block filter */
 	mlfi_eom,	/* end of message */
 	mlfi_abort,	/* message aborted */
 	mlfi_close,	/* connection cleanup */
@@ -729,7 +730,7 @@ static struct smfiDesc smfilter =
 	NULL,   /* unknown command handler */
 	NULL,   /* data handler */
 	mlfi_negotiate,   /* negotiate */
-	
+
 	child_start,   /* child start */
 	NULL,   /* child finish */
 	NULL,   /* idle callback */
@@ -792,15 +793,11 @@ milter_settimeout(time_t t)
 void
 milter_enable_state(enum smtp_state state)
 {
-	if (state > smtp_state_first && state < smtp_state_end
-	    && state < first_used_state)
-		first_used_state = state;
-	
 	switch (state) {
 	case smtp_state_connect:
 		smfilter.xxfi_connect = mlfi_connect;
 		break;
-		
+
 	case smtp_state_helo:
 		smfilter.xxfi_helo = mlfi_helo;
 		break;
@@ -808,7 +805,7 @@ milter_enable_state(enum smtp_state state)
 	case smtp_state_envfrom:
 		smfilter.xxfi_envfrom = mlfi_envfrom;
 		break;
-		
+
 	case smtp_state_envrcpt:
 		smfilter.xxfi_envrcpt = mlfi_envrcpt;
 		break;
@@ -816,7 +813,7 @@ milter_enable_state(enum smtp_state state)
 	case smtp_state_data:
 		smfilter.xxfi_data = mlfi_data;
 		break;
-		
+
 	case smtp_state_header:
 		smfilter.xxfi_header = mlfi_header;
 		break;
@@ -824,7 +821,7 @@ milter_enable_state(enum smtp_state state)
 	case smtp_state_eoh:
 		smfilter.xxfi_eoh = mlfi_eoh;
 		break;
-		
+
 	case smtp_state_body:
 		smfilter.xxfi_body = mlfi_body;
 		break;
@@ -833,4 +830,3 @@ milter_enable_state(enum smtp_state state)
 		break;
 	}
 }
-
