@@ -1,5 +1,5 @@
 /* This file is part of Mailfromd.
-   Copyright (C) 2006-2020 Sergey Poznyakoff
+   Copyright (C) 2006-2024 Sergey Poznyakoff
 
    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
@@ -261,10 +261,11 @@ struct exception_context {
 
 struct environ_locus {
 	unsigned long file;   /* File name */
-	unsigned int line;   /* Line number */
+	unsigned int line;    /* Line number */
 };
 
 struct environ_cleanup_closure {
+	int when;
 	void (*cleanup)(void *);
 	void *data;
 };
@@ -286,6 +287,7 @@ struct eval_environ {
 	struct environ_locus locus;
 	
 	/* Temporary heap space */
+	int temp_active;
 	size_t temp_start;
 	size_t temp_size;
 
@@ -295,6 +297,7 @@ struct eval_environ {
 	
 	/* Sendmail interaction data: */
 	SMFICTX *ctx;          /* Milter Context */
+	int conn_family;       /* SMTP connection family. */
 	void *data;            /* MTA symbol table */
 	/* methods to access the latter */
 	const char *(*getsym)(void *data, const char *str);
@@ -328,8 +331,6 @@ struct eval_environ {
 	/* Function clean-up sequence */
 	mu_list_t function_cleanup_list;
 
-	mu_list_t environ_cleanup_list;
-	
 	/* Exit information */
 	sfsistat status;       /* Program exit status */
 	jmp_buf x_jmp;         /* Return point for runtime errors */
@@ -448,15 +449,7 @@ _cleanup_compare(const void *a, const void *b)
 {
 	struct environ_cleanup_closure const *closa = a;
 	struct environ_cleanup_closure const *closb = b;
-	return closa->data != closb->data;
-}
-	
-static void
-_cleanup_destroy(void *item)
-{
-	struct environ_cleanup_closure *clos = item;
-	if (clos->cleanup)
-		clos->cleanup(clos->data);
+	return !(closa->when == closb->when && closa->data == closb->data);
 }
 
 static void
@@ -464,39 +457,49 @@ env_cleanup_list_create(mu_list_t *plist)
 {
 	mu_list_t list;
 	mu_list_create(&list);//FIXME: error checking
-	mu_list_set_destroy_item(list, _cleanup_destroy);
 	mu_list_set_comparator(list, _cleanup_compare);
 	*plist = list;
 }
 
 static void
-env_cleanup_list_append(mu_list_t list, void *data, void (*func)(void *))
+env_cleanup_list_append(mu_list_t list, int when, void *data, void (*func)(void *))
 {
 	struct environ_cleanup_closure *clos = mu_zalloc(sizeof(*clos));
 
+	clos->when = when;
 	clos->cleanup = func ? func : free;
 	clos->data = data;
 	mu_list_append(list, clos);
 }
 
-void
-env_function_cleanup_flush(eval_environ_t env, void *ptr)
+static int
+_cleanup_flush (void *item, void *data)
 {
-	if (ptr) {
-		struct environ_cleanup_closure clos;
-		clos.data = ptr;
-		mu_list_remove(env->function_cleanup_list, &clos);
-	} else {
-		mu_list_clear(env->function_cleanup_list);
+	struct environ_cleanup_closure *clos = item;
+	int when = *(int*)data;
+
+	if (when == CLEANUP_ALWAYS ||
+	    clos->when == CLEANUP_ALWAYS ||
+	    clos->when == when) {
+		clos->cleanup(clos->data);
 	}
+	return 0;
+}
+
+void
+env_function_cleanup_flush(eval_environ_t env, int when)
+{
+	mu_list_foreach(env->function_cleanup_list, _cleanup_flush, &when);
+	mu_list_clear(env->function_cleanup_list);
 }			
 
 void
-env_function_cleanup_del(eval_environ_t env, void *ptr)
+env_function_cleanup_del(eval_environ_t env, int when, void *ptr)
 {
 	int rc;
 	struct environ_cleanup_closure clos, *cptr;
 
+	clos.when = when;
 	clos.data = ptr;
 
 	rc = mu_list_locate(env->function_cleanup_list, &clos, (void**)&cptr);
@@ -507,15 +510,9 @@ env_function_cleanup_del(eval_environ_t env, void *ptr)
 }
 
 void
-env_function_cleanup_add(eval_environ_t env, void *data, void (*func)(void *))
+env_function_cleanup_add(eval_environ_t env, int when, void *data, void (*func)(void *))
 {
-	env_cleanup_list_append(env->function_cleanup_list, data, func);
-}
-
-void
-env_environ_cleanup_add(eval_environ_t env, void *data, void (*func)(void *))
-{
-	env_cleanup_list_append(env->environ_cleanup_list, data, func);
+	env_cleanup_list_append(env->function_cleanup_list, when, data, func);
 }
 
 /* A call to expand_dataseg (see below) invalidates any C variables that
@@ -635,51 +632,96 @@ prog_trace(eval_environ_t env, const char *fmt, ...)
 	       buf);
 }	
 
+void instr_builtin(eval_environ_t env);
 void instr_funcall(eval_environ_t env);
 void instr_locus(eval_environ_t env);
+void instr_catch(eval_environ_t env);
+
+static struct mu_locus_point *
+stack_find_locus(eval_environ_t env, prog_counter_t pc, prog_counter_t off,
+		 struct mu_locus_point *pt)
+{
+	int i;
+
+	for (i = 0; i < 32; i++, off++) {
+		if (pc > off && prog[pc - off].c_instr == instr_locus) {
+			pt->mu_file = (char*)(env->dataseg
+				 + mf_cell_c_value(prog[pc - off + 1], ulong));
+			pt->mu_line = mf_cell_c_value(prog[pc - off + 2],
+						      uint);
+			return pt;
+		}
+	}
+	return NULL;
+}
+
+static void
+stack_trace_entry(eval_environ_t env, prog_counter_t pc,
+		  mfd_stack_framer_t tracer, void *data)
+{
+	char *name;
+	struct mu_locus_point *pt, ptbuf;
+	
+	if (mf_cell_instr(prog[pc-2]) == instr_funcall ||
+	    mf_cell_instr(prog[pc-2]) == instr_builtin) {
+		name = (char*)(env->dataseg +
+			       mf_cell_c_value(prog[pc-1], size));
+		pc -= 2;
+		pt = stack_find_locus(env, pc, 3, &ptbuf);
+	} else if (mf_cell_instr(prog[pc-2]) == instr_catch) {
+		name = "(in catch)";
+		pc -= 2;
+		pt = stack_find_locus(env, pc, 3, &ptbuf);
+	} else {
+		name = "(unknown)";
+		pt = stack_find_locus(env, pc, 0, &ptbuf);
+	}
+
+	if (pt)
+		tracer(pc, name, pt->mu_file, pt->mu_line, data);
+	else
+		tracer(pc, name, NULL, 0, data);
+}
+
+void
+env_stack_trace(eval_environ_t env, mfd_stack_framer_t framer, void *data)
+{
+	size_t base;
+
+	stack_trace_entry(env, env->pc, framer, data);
+	for (base = env->base; base < datasize + env->stack_size - 4;
+	     base = mf_c_val(env->dataseg[base + 1], size) + base + 1) {
+		prog_counter_t pc = mf_c_val(env->dataseg[base + 2], ulong);
+		if (pc < 2)
+			break; /*FIXME*/
+		stack_trace_entry(env, pc, framer, data);
+		if (mf_cell_instr(prog[pc-2]) == instr_catch) {
+			stack_trace_entry(env,
+					  mf_c_val(env->dataseg[base + 5],
+						   size),
+					  framer, data);
+		}
+	}
+}
+
+static void
+print_frame(prog_counter_t pc, char const *func_name,
+	    char const *file_name, unsigned line_num, void *data)
+{
+	if (file_name)
+		mu_error("%04lu: %s:%u: %s",
+			 (unsigned long) pc,
+			 file_name, line_num, func_name);
+	else
+		mu_error("%04lu: %s",
+			 (unsigned long) pc, func_name);
+}
 
 void
 runtime_stack_trace(eval_environ_t env)
 {
-	size_t base;
-	
 	mu_error(_("stack trace:"));
-	for (base = env->base; base < datasize + env->stack_size - 4;
-	     base = mf_c_val(env->dataseg[base + 1], size) + base + 1) {
-		int i;
-		prog_counter_t pc = mf_c_val(env->dataseg[base + 2], ulong) - 1;
-		char *name;
-		struct mu_locus_point *ploc = NULL, loc;
-
-		if (pc < 2)
-			break; /*FIXME*/
-		if (mf_cell_instr(prog[pc-2]) == instr_funcall) {
-			name = (char*)(env->dataseg + mf_cell_c_value(prog[pc-1], size));
-			pc -= 2;
-		} else {
-			name = "(in catch)";
-			pc -= 3;
-		}
-
-		for (i = 0; i < 10; i++) {
-			if (pc > i + 3
-			    && prog[pc - i - 3].c_instr == instr_locus) {
-				loc.mu_file = (char*)(env->dataseg
-						   + mf_cell_c_value(prog[pc - i - 2], size));
-				loc.mu_line = mf_cell_c_value(prog[pc - i - 1], size);
-				ploc = &loc;
-				break;
-			}
-		}
-		
-		if (ploc)
-			mu_error("%04lu: %s:%u: %s",
-				 (unsigned long) pc,
-				 ploc->mu_file, ploc->mu_line, name);
-		else
-			mu_error("%04lu: %s",
-				 (unsigned long) pc, name);
-	}
+	env_stack_trace(env, print_frame, NULL);
 	mu_error(_("stack trace finishes"));
 }
 
@@ -692,7 +734,7 @@ runtime_warning(eval_environ_t env, const char *fmt, ...)
 		       _("RUNTIME WARNING near %s:%u: "),
 		       ENV_LOC_FILE(env), ENV_LOC_LINE(env));
 	va_start(ap, fmt);
-	mu_diag_cont_printf(fmt, ap);
+	mu_diag_cont_vprintf(fmt, ap);
 	va_end(ap);
 	mu_diag_cont_printf("\n");
 }
@@ -818,14 +860,18 @@ heap_obstack_begin(eval_environ_t env)
 {
 	env->temp_start = env->toh;
 	env->temp_size = 0;
+	env->temp_active = 1;
 }
 
 void
 heap_obstack_cancel(eval_environ_t env)
 {
-	env->toh = env->temp_start;
-	env->temp_start = 0;
-	env->temp_size = 0;
+	if (env->temp_active) {
+		env->toh = env->temp_start;
+		env->temp_start = 0;
+		env->temp_size = 0;
+		env->temp_active = 0;
+	}
 }
 
 STKVAL
@@ -835,6 +881,7 @@ heap_obstack_finish(eval_environ_t env)
 	mf_c_val(ret, size) = env->temp_start;
 	env->temp_start = 0;
 	env->temp_size = 0;
+	env->temp_active = 0;
 	return ret;
 }
 
@@ -907,12 +954,38 @@ heap_obstack_grow(eval_environ_t env, void * MFL_DATASEG ptr, size_t size)
 	return ret;
 }
 
+void
+heap_obstack_reclaim(eval_environ_t env, size_t size)
+{
+	if (size > env->temp_size)
+		abort();
+	env->temp_size -= size;
+	env->toh -= B2STACK(size);
+}
+
+void
+heap_obstack_truncate(eval_environ_t env, size_t size)
+{
+	if (size == env->temp_size)
+		return;
+	else if (size > env->temp_size)
+		heap_obstack_grow(env, NULL, size - env->temp_size);
+	else
+		heap_obstack_reclaim(env, env->temp_size - size);
+}
+
 void *
 heap_obstack_base(eval_environ_t env)
 {
 	return (void*) env_data_ref(env, env->temp_start);
 }
 
+size_t
+heap_obstack_length(eval_environ_t env)
+{
+	return env->temp_size;
+}
+
 STKVAL *
 env_data_ref(eval_environ_t env, size_t off)
 {
@@ -942,7 +1015,7 @@ pushn(eval_environ_t env, long n)
 void
 instr_locus(eval_environ_t env)
 {
-	env->locus.file = mf_c_val(get_immediate(env, 0), uint);
+	env->locus.file = mf_c_val(get_immediate(env, 0), ulong);
 	env->locus.line = mf_c_val(get_immediate(env, 1), uint);
 	if (PROG_TRACE_ENGINE)
 		prog_trace(env, "LOCUS");
@@ -953,8 +1026,8 @@ void
 dump_locus(prog_counter_t i)
 {
 	printf("\"%s\" %lu",
-	       (char*) (dataseg + mf_cell_c_value(prog[i], size)),
-	       (unsigned long) mf_cell_c_value(prog[i+1], size));
+	       (char*) (dataseg + mf_cell_c_value(prog[i], ulong)),
+	       (unsigned long) mf_cell_c_value(prog[i+1], uint));
 }
 
 /* Stack manipulation instructions */
@@ -972,9 +1045,42 @@ instr_xchg(eval_environ_t env)
 void
 instr_dup(eval_environ_t env)
 {
-	if (PROG_TRACE_ENGINE)
+        if (PROG_TRACE_ENGINE)
 		prog_trace(env, "DUP");
-	push(env, get_arg(env, 0));
+	push(env, (STKVAL)0L);
+	env->dataseg[env->tos+1] = env->dataseg[env->tos+2];
+}
+
+/*
+ * push COUNT
+ * push OFFSET
+ * dupn FRAME
+ *
+ * Pushes on stack COUNT stack words starting from OFFSET in the given
+ * FRAME, in the same order they appear on stack.
+ */
+void
+instr_dupn(eval_environ_t env)
+{
+	size_t frame = mf_c_val(get_immediate(env, 0), size);
+	long count = mf_c_val(get_arg(env, 1), long);
+	long off = mf_c_val(get_arg(env, 0), long);
+	size_t i;
+
+	if (count < 0)
+		runtime_error(env, _("dupn: negative count"));
+	if (PROG_TRACE_ENGINE)
+		prog_trace(env, "DUPN frame=%zu off=%ld count=%zu",
+			   frame, off, count);
+
+	adjust_stack(env, 2);
+
+	for (i = count; i > 0; i--) {
+		push(env, (STKVAL)0L);
+		env->dataseg[env->tos+1] = env->dataseg[env_base(env, frame) + off + i - 1];
+	}
+	
+	advance_pc(env, 1);
 }
 
 void
@@ -1013,7 +1119,7 @@ instr_memstk(eval_environ_t env)
 	      push(env, env_base(env, frame) + off)
 
 	   However, push() itself might cause dataseg expansion,
-	   which would change tos/toh pointers and render the previously
+	   which would change tos/base pointers and render the previously
 	   computed argument value invalid.  The correct way is to push
 	   a placeholder value first, thereby occupying stack slot and
 	   eventually causing dataseg expansion, and then compute val
@@ -1090,6 +1196,7 @@ instr_stkalloc(eval_environ_t env)
 				      _("Out of stack space; increase #pragma stacksize"));
 	}
 	env->tos -= n;
+	memset(env->dataseg + env->tos, 0, n * sizeof(env->dataseg[0]));
 	advance_pc(env, 1);
 }
 
@@ -1224,7 +1331,7 @@ dump_symbol(prog_counter_t i)
 	       (char *) (dataseg + mf_cell_c_value(prog[i], size)));
 }
 
-/* Comparation instructions */
+/* Comparison instructions */
 void
 instr_eqn(eval_environ_t env)
 {
@@ -1414,7 +1521,7 @@ instr_jmp(eval_environ_t env)
 	advance_pc(env, off+1);
 }
 
-/* Longean instructions */
+/* Boolean instructions */
 void
 instr_not(eval_environ_t env)
 {
@@ -1425,7 +1532,7 @@ instr_not(eval_environ_t env)
 	pushn(env, !v);
 }
 
-/* Bitwise arithmetics */
+/* Bitwise arithmetic */
 void
 instr_logand(eval_environ_t env)
 {
@@ -1854,7 +1961,7 @@ instr_result(eval_environ_t env)
 			   SP(code),
 			   SP(xcode),
 			   SP(message));
-	
+
 	trace("%s%s:%u: %s %s %s %s",
 	      mailfromd_msgid(env->ctx),
 	      ENV_LOC_FILE(env), ENV_LOC_LINE(env),
@@ -1873,6 +1980,7 @@ instr_result(eval_environ_t env)
 				       "modification commands to be ignored; "
 				       "call mmq_purge() prior to `accept', "
 				       "to suppress this warning"));
+
 	env->status = status;
 	env->setreply(env->data, code, xcode, message);
 	advance_pc(env, 1);
@@ -1955,12 +2063,6 @@ instr_concat(eval_environ_t env)
 	push(env, (STKVAL) off);
 }
 
-void
-dump_adjust(prog_counter_t i)
-{
-	printf("%lu ", mf_cell_c_value(prog[i], ulong));
-}
-
 void
 instr_asgn(eval_environ_t env)
 {
@@ -1968,9 +2070,9 @@ instr_asgn(eval_environ_t env)
 	size_t dest = mf_c_val(get_arg(env, 0), size);
 	adjust_stack(env, 2);
 	if (PROG_TRACE_ENGINE) 
-		prog_trace(env, "ASGN %lu=%u",
+		prog_trace(env, "ASGN %lu=%lu",
 			   (unsigned long) dest,
-			   mf_c_val(val, uint));
+			   mf_c_val(val, ulong));
 	env->dataseg[dest] = val;
 }
 
@@ -1978,7 +2080,7 @@ void
 instr_catch(eval_environ_t env)
 {
 	long off = mf_c_val(get_immediate(env, 0), long);
-	unsigned toff = mf_c_val(get_immediate(env, 1), uint);
+	size_t toff = mf_c_val(get_immediate(env, 1), size);
 	size_t count = mf_c_val(env->dataseg[toff], size);
 	STKVAL *tab = (STKVAL *) (env->dataseg + toff + 1);
 	size_t i;
@@ -2023,7 +2125,6 @@ instr_throw(eval_environ_t env)
 {
 	unsigned long n = mf_c_val(get_immediate(env, 0), ulong);
 	size_t off = mf_c_val(get_arg(env, 0), size);
-	advance_pc(env, 1);
 	adjust_stack(env, 1);
 	if (n > exception_count)
 		runtime_error(env, _("invalid exception number: %lu"), n);
@@ -2035,8 +2136,8 @@ instr_throw(eval_environ_t env)
 void
 dump_throw(prog_counter_t i)
 {
-	printf("%s (%u)", mf_exception_str(mf_cell_c_value(prog[i], uint)),
-	       mf_cell_c_value(prog[i], uint)); 
+	printf("%s (%lu)", mf_exception_str(mf_cell_c_value(prog[i], ulong)),
+	       mf_cell_c_value(prog[i], ulong)); 
 }
 
 
@@ -2058,7 +2159,6 @@ instr_return(eval_environ_t env)
 	if (PROG_TRACE_ENGINE)
 		prog_trace(env, "RETURN");
 	env_leave_frame(env, 0);
-	env->pc--;
 }
 
 void
@@ -2067,25 +2167,40 @@ instr_retcatch(eval_environ_t env)
 	prog_counter_t pc = env->pc;
 	if (PROG_TRACE_ENGINE)
 		prog_trace(env, "RETCATCH");
-	env_leave_frame(env, 2);
+	env_leave_frame(env, 3);
 	env->pc = pc;
 }
 
+/*
+ * Return true if the given bit is set in the exception mask starting at
+ * offset off.
+ */
+static inline int
+env_exmask_isset(eval_environ_t env, size_t off, int bit)
+{
+	return ((bitmask_bits_t)mf_c_val(env->dataseg[off + 1 + BIT_WORD(bit)], uint)) & BIT_MASK(bit);
+}
+
 void
 instr_saveex(eval_environ_t env)
 {
-	unsigned off = mf_c_val(get_immediate(env, 0), uint);
+	size_t off = mf_c_val(get_immediate(env, 0), size);
 	size_t count = mf_c_val(env->dataseg[off], size);
-	STKVAL *tab = (STKVAL *) (env->dataseg + off + 1);
 	size_t i;
 
 	if (PROG_TRACE_ENGINE)
-		prog_trace(env, "SAVEEX %x (%lu ex.)", off,
+		prog_trace(env, "SAVEEX %lx (%lu ex.)",
+			   (unsigned long) off,
 			   (unsigned long) count);
 
 	advance_pc(env, 1);
+	/*
+	 * Note: in the loop that follows each push can lead to data segment
+	 * expansion, therefore the actual start address of the exception table
+	 * needs to be computed in each iteration using env_exmask_isset.
+	 */
 	for (i = 0; i < count * NBMBITS; i++)
-		if (((bitmask_bits_t)mf_c_val(tab[BIT_WORD(i)], uint)) & BIT_MASK(i)) {
+		if (env_exmask_isset(env, off, i)) {
 			mu_debug(MF_SOURCE_PROG, MU_DEBUG_TRACE9,
 			      ("Push Exception: %lu %lu <- pc=%lu, tos=%lu, base=%lu",
 			       (unsigned long) i,
@@ -2117,14 +2232,14 @@ dump_saveex(prog_counter_t ctr)
 void
 instr_restex(eval_environ_t env)
 {
-	unsigned off = mf_c_val(pop(env), uint);
+	size_t off = mf_c_val(pop(env), size);
 	size_t count = mf_c_val(env->dataseg[off], size);
 	STKVAL *tab = (STKVAL *) (env->dataseg + off + 1);
 	size_t i;
 	
 	if (PROG_TRACE_ENGINE)
-		prog_trace(env, "RESTEX %x (%lu ex.)",
-			   off, (unsigned long) count);
+		prog_trace(env, "RESTEX %lx (%lu ex.)",
+			   (unsigned long) off, (unsigned long) count);
 	if (!count)
 		return;
 	i = count * NBMBITS - 1;
@@ -2145,6 +2260,12 @@ instr_restex(eval_environ_t env)
 	} while (i > 0);
 }
 
+void
+dump_adjust(prog_counter_t i)
+{
+	printf("%lu ", mf_cell_c_value(prog[i], ulong));
+}
+
 void
 instr_adjust(eval_environ_t env)
 {
@@ -2155,6 +2276,30 @@ instr_adjust(eval_environ_t env)
 	advance_pc(env, 1);
 }
 
+/*
+ * Adjust stack after return from a variadic function.
+ *
+ * When such function is called, it is passed the actual number
+ * of arguments in its first (hidden) argument.
+ *
+ * The inline parameter to adjustx gives the number of arguments the
+ * function is declared to take (both mandatory and optional).  These
+ * are always on stack, so this value is used instead of the actual
+ * number if the latter is less than the former.  This can happen,
+ * e.g. if a function takes both optional and variable arguments and
+ * is given only mandatory ones.
+ */
+void
+instr_adjustx(eval_environ_t env)
+{
+	long nargs = mf_c_val(get_immediate(env, 0), long);
+	size_t n = mf_c_val(pop(env), size);
+	if (PROG_TRACE_ENGINE)
+		prog_trace(env, "ADJUSTX %zu %ld", n, nargs);
+	adjust_stack(env, n < nargs ? nargs : n);
+	advance_pc(env, 1);
+}
+
 void
 instr_popreg(eval_environ_t env)
 {
@@ -2215,44 +2360,45 @@ instr_xlat(eval_environ_t env)
 	for (i = 0; i < count; i += 2) {
 		if (mf_c_val(tab[i], long) == mf_c_val(env->reg, long)) {
 			env->reg = tab[i+1];
-			push(env, (STKVAL)0);
+			push(env, (STKVAL)0L);
 			return;
 		}
 	}
-	push(env, (STKVAL)1);
+	push(env, (STKVAL)1L);
 }
 
 void
 instr_xlats(eval_environ_t env)
 {
-	unsigned long i;
-	unsigned long count = mf_c_val(get_immediate(env, 0), uint);
+	size_t i;
+	size_t count = mf_c_val(get_immediate(env, 0), size);
 	size_t off = mf_c_val(get_immediate(env, 1), size);
 	STKVAL *tab = (STKVAL *) (env->dataseg + off);
 	char *str = (char*) env_data_ref(env, mf_c_val(env->reg, size));
 	
         if (PROG_TRACE_ENGINE)
-		prog_trace(env, "XLATS %lu %lu", count, (unsigned long) off);
+		prog_trace(env, "XLATS %lu %lu",
+			   (unsigned long) count, (unsigned long) off);
 	advance_pc(env, 2);
 	for (i = 0; i < count; i += 2) {
 		if (strcmp((char*)(env->dataseg + mf_c_val(tab[i], size)), str) == 0) {
 			env->reg = tab[i+1];
-			push(env, (STKVAL)0);
+			push(env, (STKVAL)0L);
 			return;
 		}
 	}
-	push(env, (STKVAL)1);
+	push(env, (STKVAL)1L);
 }
 
 void
 dump_xlat(prog_counter_t i)
 {
-	unsigned long j;
-	unsigned long count = mf_cell_c_value(prog[i], ulong);
+	size_t j;
+	size_t count = mf_cell_c_value(prog[i], size);
 	size_t off = mf_cell_c_value(prog[i+1], size);
 	STKVAL *tab = (STKVAL *) (dataseg + off);
 	
-	printf("%lu %lu ", count, (unsigned long) off);
+	printf("%lu %lu ", (unsigned long) count, (unsigned long) off);
 	for (j = 0; j < count; j += 2) 
 		printf("(%ld %ld) ", mf_c_val(tab[j], long), mf_c_val(tab[j+1], long));
 }
@@ -2354,6 +2500,7 @@ env_init(eval_environ_t env)
 	env->status = SMFIS_CONTINUE;
 	env->tos = datasize + env->stack_size - 1;
 	env->base = 0;
+	env->pc = 0;
 	mf_c_val(env->reg, long) = 0; //FIXME
 
 	env->matches = NULL;
@@ -2384,18 +2531,10 @@ env_init_dataseg(eval_environ_t env)
 	init_dataseg(env->dataseg, dvarsize);
 }
 
-void
-env_make_frame0(eval_environ_t env)
-{
-	push(env, (STKVAL) 0);
-	push(env, (STKVAL) (env->base - env->tos));
-	env->base = env->tos;
-}
-
 void
 env_make_frame(eval_environ_t env)
 {
-	push(env, (STKVAL) (env->pc + 1));
+	push(env, (STKVAL) env->pc);
 	push(env, (STKVAL) (env->base - env->tos));
 	env->base = env->tos;
 }
@@ -2510,13 +2649,15 @@ env_throw_0(eval_environ_t env, mf_exception status, size_t off)
 {
 	prog_counter_t pc;
 
-	env_function_cleanup_flush(env, NULL);
+	env_function_cleanup_flush(env, CLEANUP_THROW);
 	if (status > exception_count)
 		runtime_error(env, _("unknown exception: %d: %s"),
 			      status, (char*) env_data_ref(env, off));
 	
 	pc = env->catch_ctx[status].pc;
 	if (pc) {
+		prog_counter_t throw_pc = env->pc;
+
 		/* Restore tos */
 		env->tos = TOS_INVARIANT(env, env->catch_ctx[status].tos);
 		env->base = TOS_INVARIANT(env, env->catch_ctx[status].base);
@@ -2525,6 +2666,7 @@ env_throw_0(eval_environ_t env, mf_exception status, size_t off)
 		/* Fixup the program counter */
 		env->pc = pc - 1;
 		/* Generate normal entry frame */
+		push(env, (STKVAL) throw_pc);
 		push(env, (STKVAL) off); 
 		env_push_number(env, status);
 		env_make_frame(env);
@@ -2552,6 +2694,9 @@ env_throw_bi(eval_environ_t env, mf_exception status, const char *biname,
 {
 	va_list ap;
 	size_t off;
+
+	heap_obstack_cancel(env);
+	
 	va_start(ap, fmt);
 	off = env_vsprintf(env, biname, fmt, ap);
 	va_end(ap);
@@ -2599,7 +2744,7 @@ env_capture_start(eval_environ_t env)
 		/* Drop any previously captured message data */
 		env_free_captured(env);
 		mu_header_destroy(&env->header);
-		/* Truncate existing stream and reposition to its
+		/* Truncate existing stream and re-position to its
 		   beginning */
 		rc = mu_stream_truncate(env->stream, 0);
 		if (rc == 0 &&
@@ -2741,7 +2886,7 @@ env_get_header(eval_environ_t env, mu_header_t *hdr)
 		}
 		text[total] = 0;
 
-               /* FIXME: Reposition the stream at its end.
+               /* FIXME: Re-position the stream at its end.
 		  This call may happen in the middle of capturing
 		  so I have to make sure to not disturb the capturing
 		  process.
@@ -2923,6 +3068,18 @@ env_builtin_priv_destroy(eval_environ_t env)
 	}
 }
 
+void
+env_save_socket_family(eval_environ_t env, int family)
+{
+	env->conn_family = family;
+}
+
+int
+env_socket_family(eval_environ_t env)
+{
+	return env->conn_family;
+}
+
 eval_environ_t
 create_environment(SMFICTX *ctx,
 		   const char *(*getsym)(void *data, const char *str),
@@ -2948,6 +3105,11 @@ create_environment(SMFICTX *ctx,
 	init_dataseg(env->dataseg, datasize);
 	
 	env->ctx = ctx;
+	/*
+	 * Assume inet address family.  It will be replaced with the
+	 * actual value when mlfi_connect is called.
+	 */
+	env->conn_family = AF_INET;
 	env->data = data;
 	env->getsym = getsym;
 	env->setreply = setreply;
@@ -2964,7 +3126,6 @@ create_environment(SMFICTX *ctx,
 	env->bi_priv_array = NULL;
 
 	env_cleanup_list_create(&env->function_cleanup_list);
-	env_cleanup_list_create(&env->environ_cleanup_list);
 	
 	if (exception_count) {
 		env->defcatch_ctx = mu_zalloc(exception_count *
@@ -2986,11 +3147,67 @@ destroy_environment(eval_environ_t env)
 	mu_stream_destroy(&env->stream);
 	mu_header_destroy(&env->header);
 	mu_list_destroy(&env->function_cleanup_list);
-	mu_list_destroy(&env->environ_cleanup_list);
 	mu_list_destroy(&env->mmq);
 	env_builtin_priv_destroy(env);
 	free(env);
 }
+
+static const char *
+null_getsym(void *data, const char *str)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of $macro in startup handler"));
+}
+
+static int
+null_setreply(void *data, char *code, char *xcode, char *message)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of milter action in startup handler"));
+}
+static void
+null_msgmod(void *data, struct msgmod_closure *clos)
+{
+	eval_environ_t *env = data;
+	runtime_error(*env, "%s", _("invalid use of message modification function in startup handler"));
+}
+
+void
+run_shutdown(void)
+{
+	if (entry_point[smtp_state_shutdown]) {
+		eval_environ_t env;
+		env = create_environment(NULL,
+					 null_getsym,
+					 null_setreply,
+					 null_msgmod,
+					 &env);
+		env_init(env);
+		xeval(env, smtp_state_shutdown);
+		destroy_environment(env);
+	}
+	mu_stream_close(mf_strecho);
+}
+
+void
+run_startup(void)
+{
+	if (entry_point[smtp_state_startup]) {
+		eval_environ_t env;
+		env = create_environment(NULL,
+					 null_getsym,
+					 null_setreply,
+					 null_msgmod,
+					 &env);
+		env_init(env);
+		if (xeval(env, smtp_state_startup))
+			exit(EX_CONFIG);
+		env_final_gc(env);
+		datasize = env->toh;
+		memcpy(dataseg, env->dataseg, datasize * sizeof dataseg[0]);
+		destroy_environment(env);
+	}
+}
 
 
 struct entry_point {
@@ -3086,7 +3303,7 @@ print_code()
 				 function_counter, &d);
 	}
 	epcount = d.i;
-	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
+	for (tag = smtp_state_startup; tag < smtp_state_count; tag++)
 		if (entry_point[tag])
 			epcount++;
 	
@@ -3094,7 +3311,7 @@ print_code()
 
         /* Fill in entry points array */
 	i = 0;
-	for (tag = smtp_state_first; tag < smtp_state_count; tag++)
+	for (tag = smtp_state_startup; tag < smtp_state_count; tag++)
 		if (entry_point[tag]) {
 			ep[i].ishandler = 1;
 			ep[i].pc = entry_point[tag];
@@ -3179,8 +3396,9 @@ env_final_gc(eval_environ_t env)
 			env->dataseg[off] = (STKVAL) q;
 #undef S_PTR
 		}
-		free(s_off);
 	}
+
+	free(s_off);
 }
 
 void
